"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Socket connection
"""
import time
import threading
import logging
from queue import Queue
import socket
from plugwise.constants import SLEEP_TIME
from plugwise.connections.connection import StickConnection
from plugwise.message import PlugwiseMessage
from plugwise.util import PlugwiseException


class SocketConnection(StickConnection):
    """
    Wrapper for Socket connection configuration
    """

    def __init__(self, device, stick=None):
        StickConnection.__init__(self)
        self.logger = logging.getLogger("plugwise")
        self._device = device
        self.stick = stick
        # get the address from a <host>:<port> format
        addr = device.split(":")
        addr = (addr[0], int(addr[1]))
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect(addr)
        except Exception:
            self.logger.error(
                "Could not open socket, \
                              no messages are read or written to the bus"
            )
            raise plugwiseException("Could not open socket port")
        # build a read thread
        self._listen_process = threading.Thread(
            None, self.read_daemon, "plugwise-process-reader", (), {}
        )
        self._listen_process.daemon = True
        self._listen_process.start()

        # build a writer thread
        self._write_queue = Queue()
        self._write_process = threading.Thread(
            None, self.write_daemon, "plugwise-connection-writer", (), {}
        )
        self._write_process.daemon = True
        self._write_process.start()

    def stop_connection(self):
        """Close the socket."""
        self.logger.warning("Stop executed")
        try:
            self._socket.close()
        except Exception:
            self.logger.error("Error while closing socket")
            raise plugwiseException("Error while closing socket")
        time.sleep(1)

    def feed_parser(self, data):
        """Parse received message."""
        assert isinstance(data, bytes)
        self.stick.feed_parser(data)

    def send(self, message, callback=None):
        """Add message to write queue."""
        assert isinstance(message, PlugwiseMessage)
        self._write_queue.put_nowait((message, callback))

    def read_daemon(self):
        """Read thread."""
        while True:
            data = self._socket.recv(9999)
            self.feed_parser(data)

    def write_daemon(self):
        """Write thread."""
        while True:
            (message, callback) = self._write_queue.get(block=True)
            self.logger.info("Sending message on USB bus: %s", str(message))
            self.logger.error("Sending binary message:  %s", str(message.serialize()))
            self._socket.send(message.serialize())
            time.sleep(SLEEP_TIME)
            if callback:
                callback()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Circle+ node object
"""
from datetime import datetime
from plugwise.constants import MAX_TIME_DRIFT
from plugwise.node import PlugwiseNode
from plugwise.nodes.circle import PlugwiseCircle
from plugwise.messages.requests import (
    CirclePlusRealTimeClockGetRequest,
    CirclePlusRealTimeClockSetRequest,
    CirclePlusScanRequest,
)
from plugwise.messages.responses import (
    CirclePlusRealTimeClockResponse,
    CircleScanResponse,
)


class PlugwiseCirclePlus(PlugwiseCircle):
    """provides interface to the Plugwise Circle+ nodes
    """

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self._plugwise_nodes = []
        self._scan_for_nodes_callback = None
        self._print_progress = False
        self._realtime_clock_offset = None
        self.get_real_time_clock(self.sync_realtime_clock)

    def _circle_plus_message(self, message):
        """
        Process received message
        """
        if isinstance(message, CirclePlusRealTimeClockResponse):
            self._response_realtime_clock(message)
        elif isinstance(message, CircleScanResponse):
            self._process_scan_response(message)
        else:
            self.stick.logger.waning(
                "Unsupported message type '%s' received from circle with mac %s",
                str(message.__class__.__name__),
                self.get_mac(),
            )
        self.stick.message_processed(message.seq_id)

    def scan_for_nodes(self, callback=None):
        self._scan_for_nodes_callback = callback
        for node_address in range(0, 64):
            self.stick.send(CirclePlusScanRequest(self.mac, node_address))

    def _process_scan_response(self, message):
        """ Process scan response message """
        self.stick.logger.debug(
            "Process scan response for address %s", message.node_address.value
        )
        if message.node_mac.value != b"FFFFFFFFFFFFFFFF":
            if self.stick.print_progress:
                print(
                    "Scan at address "
                    + str(message.node_address.value)
                    + " => node found with mac "
                    + message.node_mac.value.decode("ascii")
                )
            self.stick.logger.debug(
                "Linked plugwise node with mac %s found",
                message.node_mac.value.decode("ascii"),
            )
            self._plugwise_nodes.append(
                [message.node_mac.value.decode("ascii"), message.node_address.value]
            )
        else:
            if self.stick.print_progress:
                print(
                    "Scan at address "
                    + str(message.node_address.value)
                    + " => no node found"
                )
        if message.node_address.value == 63 and self._scan_for_nodes_callback:
            self._scan_for_nodes_callback(self._plugwise_nodes)
            self._scan_for_nodes_callback = None
            self._plugwise_nodes = []

    def get_real_time_clock(self, callback=None):
        """ get current datetime of internal clock of CirclePlus """
        self.stick.send(
            CirclePlusRealTimeClockGetRequest(self.mac), callback,
        )

    def _response_realtime_clock(self, message):
        dt = datetime(
            datetime.now().year,
            datetime.now().month,
            datetime.now().day,
            message.time.value.hour,
            message.time.value.minute,
            message.time.value.second,
        )
        realtime_clock_offset = message.timestamp.replace(microsecond=0) - (
            dt + self.stick.timezone_delta
        )
        if realtime_clock_offset.days == -1:
            self._realtime_clock_offset = realtime_clock_offset.seconds - 86400
        else:
            self._realtime_clock_offset = realtime_clock_offset.seconds
        self.stick.logger.debug(
            "Realtime clock of node %s has drifted %s sec",
            self.get_mac(),
            str(self._clock_offset),
        )

    def set_real_time_clock(self, callback=None):
        """ set internal clock of CirclePlus """
        self.stick.send(
            CirclePlusRealTimeClockSetRequest(self.mac, datetime.utcnow()), callback,
        )

    def sync_realtime_clock(self, max_drift=0):
        """ Sync real time clock of node if time has drifted more than max drifted
        """
        if self._realtime_clock_offset != None:
            if max_drift == 0:
                max_drift = MAX_TIME_DRIFT
            if (self._realtime_clock_offset > max_drift) or (
                self._realtime_clock_offset < -(max_drift)
            ):
                self.stick.logger.warning(
                    "Reset realtime clock of node %s because time has drifted %s sec",
                    self.get_mac(),
                    str(self._clock_offset),
                )
                self.set_real_time_clock()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

General node object to control associated plugwise nodes like: Circle+, Circle, Scan, Stealth
"""
from datetime import datetime
from plugwise.constants import (
    HA_SWITCH,
    MAX_TIME_DRIFT,
    NODE_TYPE_CIRCLE,
    NODE_TYPE_CIRCLE_PLUS,
    NODE_TYPE_SCAN,
    NODE_TYPE_SENSE,
    NODE_TYPE_STEALTH,
    NODE_TYPE_SWITCH,
    NODE_TYPE_STICK,
    SENSOR_AVAILABLE,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SENSOR_PING,
    SWITCH_RELAY,
)
from plugwise.message import PlugwiseMessage
from plugwise.messages.responses import (
    NodeClockResponse,
    NodeInfoResponse,
    NodePingResponse,
)
from plugwise.messages.requests import (
    NodeClockGetRequest,
    NodeClockSetRequest,
    NodeInfoRequest,
    NodePingRequest,
)
from plugwise.util import validate_mac


class PlugwiseNode(object):
    """provides interface to the Plugwise node devices
    """

    def __init__(self, mac, address, stick):
        mac = mac.upper()
        if validate_mac(mac) == False:
            self.stick.logger.debug(
                "MAC address is in unexpected format: %s", str(mac),
            )
        self.mac = bytes(mac, encoding="ascii")
        self.stick = stick
        self.categories = ()
        self.sensors = ()
        self.switches = ()
        self._address = address
        self._callbacks = {}
        self.last_update = None
        self.last_request = None
        self._available = False
        self.in_RSSI = None
        self.out_RSSI = None
        self.ping_ms = None
        self._node_type = None
        self._hardware_version = None
        self._firmware_version = None
        self._relay_state = False
        self._last_log_address = None
        self._last_log_collected = False
        self._last_info_message = None
        self._clock_offset = None
        self.get_clock(self.sync_clock)

    def get_categories(self) -> tuple:
        """ Return Home Assistant catagories supported by plugwise node """
        return self.categories

    def get_sensors(self) -> tuple:
        """ Return sensors supported by plugwise node """
        return self.sensors

    def get_switches(self) -> tuple:
        """ Return switches supported by plugwise node """
        return self.switches

    def get_available(self) -> bool:
        """ Return current network state of plugwise node """
        return self._available

    def set_available(self, state, request_info=False):
        """ Set current network state of plugwise node """
        if state == True:
            if self._available == False:
                self._available = True
                self.stick.logger.debug(
                    "Mark node %s available", self.get_mac(),
                )
                self.do_callback(SENSOR_AVAILABLE["id"])
                if request_info:
                    self._request_info()
        else:
            if self._available == True:
                self._available = False
                self.stick.logger.debug(
                    "Mark node %s unavailable", self.get_mac(),
                )
                self.do_callback(SENSOR_AVAILABLE["id"])

    def get_mac(self) -> str:
        """Return mac address"""
        return self.mac.decode("ascii")

    def get_name(self) -> str:
        """Return unique name"""
        return self.get_node_type() + " (" + str(self._address) + ")"

    def get_node_type(self) -> str:
        """Return Circle type"""
        if self._node_type == NODE_TYPE_CIRCLE:
            return "Circle"
        elif self._node_type == NODE_TYPE_CIRCLE_PLUS:
            return "Circle+"
        elif self._node_type == NODE_TYPE_SCAN:
            return "Scan"
        elif self._node_type == NODE_TYPE_SENSE:
            return "Sense"
        elif self._node_type == NODE_TYPE_STEALTH:
            return "Stealth"
        elif self._node_type == NODE_TYPE_SWITCH:
            return "Switch"
        elif self._node_type == NODE_TYPE_STICK:
            return "Stick"
        return "Unknown"

    def get_hardware_version(self) -> str:
        """Return hardware version"""
        if self._hardware_version != None:
            return self._hardware_version
        return "Unknown"

    def get_firmware_version(self) -> str:
        """Return firmware version"""
        if self._firmware_version != None:
            return str(self._firmware_version)
        return "Unknown"

    def get_last_update(self) -> datetime:
        """Return  version"""
        return self.last_update

    def get_in_RSSI(self) -> int:
        """Return inbound RSSI level"""
        if self.in_RSSI != None:
            return self.in_RSSI
        return 0

    def get_out_RSSI(self) -> int:
        """Return outbound RSSI level"""
        if self.out_RSSI != None:
            return self.out_RSSI
        return 0

    def get_ping(self) -> int:
        """Return ping roundtrip"""
        if self.ping_ms != None:
            return self.ping_ms
        return 0

    def _request_info(self, callback=None):
        """ Request info from node"""
        self.stick.send(
            NodeInfoRequest(self.mac), callback,
        )

    def ping(self, callback=None):
        """ Ping node"""
        self.stick.send(
            NodePingRequest(self.mac), callback,
        )

    def on_message(self, message):
        """
        Process received message
        """
        assert isinstance(message, PlugwiseMessage)
        if message.mac == self.mac:
            if message.timestamp != None:
                self.stick.logger.debug(
                    "Last update %s of node %s, last message %s",
                    str(self.last_update),
                    self.get_mac(),
                    str(message.timestamp),
                )
                self.last_update = message.timestamp
            if isinstance(message, NodePingResponse):
                self._process_ping_response(message)
                self.stick.message_processed(message.seq_id)
            elif isinstance(message, NodeInfoResponse):
                self._process_info_response(message)
                self.stick.message_processed(message.seq_id)
            elif isinstance(message, NodeClockResponse):
                self._response_clock(message)
                self.stick.message_processed(message.seq_id)
            else:
                self.set_available(True)
                self._on_message(message)
        else:
            self.stick.logger.debug(
                "Skip message, mac of node (%s) != mac at message (%s)",
                message.mac.decode("ascii"),
                self.get_mac(),
            )

    def _on_message(self, message):
        pass

    def subscribe_callback(self, callback, sensor):
        """ Subscribe callback to execute when state change happens """
        if sensor not in self._callbacks:
            self._callbacks[sensor] = []
        self._callbacks[sensor].append(callback)

    def unsubscribe_callback(self, callback, sensor):
        """ Register callback to execute when state change happens """
        if sensor in self._callbacks:
            self._callbacks[sensor].remove(callback)

    def do_callback(self, sensor):
        """ Execute callbacks registered for specified callback type """
        if sensor in self._callbacks:
            for callback in self._callbacks[sensor]:
                try:
                    callback(None)
                except Exception as e:
                    self.stick.logger.error(
                        "Error while executing all callback : %s", e,
                    )

    def _process_ping_response(self, message):
        """ Process ping response message"""
        self.set_available(True, True)
        if self.in_RSSI != message.in_RSSI.value:
            self.in_RSSI = message.in_RSSI.value
            self.do_callback(SENSOR_RSSI_IN["id"])
        if self.out_RSSI != message.out_RSSI.value:
            self.out_RSSI = message.out_RSSI.value
            self.do_callback(SENSOR_RSSI_OUT["id"])
        if self.ping_ms != message.ping_ms.value:
            self.ping_ms = message.ping_ms.value
            self.do_callback(SENSOR_PING["id"])

    def _process_info_response(self, message):
        """ Process info response message"""
        self.stick.logger.debug("Response info message for plug %s", self.get_mac())
        self.set_available(True)
        if message.relay_state.serialize() == b"01":
            if not self._relay_state:
                self._relay_state = True
                self.do_callback(SWITCH_RELAY["id"])
        else:
            if self._relay_state:
                self._relay_state = False
                self.do_callback(SWITCH_RELAY["id"])
        self._hardware_version = int(message.hw_ver.value)
        self._firmware_version = message.fw_ver.value
        self._node_type = message.node_type.value
        self._last_info_message = message.timestamp
        if self._last_log_address != message.last_logaddr.value:
            self._last_log_address = message.last_logaddr.value
            self._last_log_collected = False
        self.stick.logger.debug("Node type        = %s", self.get_node_type())
        self.stick.logger.debug("Relay state      = %s", str(self._relay_state))
        self.stick.logger.debug("Hardware version = %s", str(self._hardware_version))
        self.stick.logger.debug("Firmware version = %s", str(self._firmware_version))

    def _request_power_buffer(self, log_address=None, callback=None):
        pass

    def get_clock(self, callback=None):
        """ get current datetime of internal clock of CirclePlus """
        self.stick.send(
            NodeClockGetRequest(self.mac), callback,
        )

    def _response_clock(self, message):
        dt = datetime(
            datetime.now().year,
            datetime.now().month,
            datetime.now().day,
            message.time.value.hour,
            message.time.value.minute,
            message.time.value.second,
        )
        clock_offset = message.timestamp.replace(microsecond=0) - (
            dt + self.stick.timezone_delta
        )
        if clock_offset.days == -1:
            self._clock_offset = clock_offset.seconds - 86400
        else:
            self._clock_offset = clock_offset.seconds
        self.stick.logger.debug(
            "Clock of node %s has drifted %s sec",
            self.get_mac(),
            str(self._clock_offset),
        )

    def set_clock(self, callback=None):
        """ set internal clock of CirclePlus """
        self.stick.send(
            NodeClockSetRequest(self.mac, datetime.utcnow()), callback,
        )

    def sync_clock(self, max_drift=0):
        """ Resync clock of node if time has drifted more than MAX_TIME_DRIFT
        """
        if self._clock_offset != None:
            if max_drift == 0:
                max_drift = MAX_TIME_DRIFT
            if (self._clock_offset > max_drift) or (self._clock_offset < -(max_drift)):
                self.stick.logger.warning(
                    "Reset clock of node %s because time has drifted %s sec",
                    self.get_mac(),
                    str(self._clock_offset),
                )
                self.set_clock()



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Base for serial or socket connections
"""

class StickConnection(object):
    """ Generic Plugwise stick connection"""

    def open_port(self) -> bool:
        """Placeholder to initialize the connection"""
        raise NotImplementedError

    def is_connected(self):
        """Placeholder to get current state of connection"""
        raise NotImplementedError

    def read_thread_alive(self):
        """Placeholder to get current state of the reader thread"""
        raise NotImplementedError

    def write_thread_alive(self):
        """Placeholder to get current state of the writer thread"""
        raise NotImplementedError

    def send(self, message, callback=None):
        """Placeholder to send message"""
        raise NotImplementedError

    def close_port(self):
        """Placeholder to disconnect"""
        raise NotImplementedError



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Serial USB connection
"""
import time
import threading
import logging
from queue import Queue
import serial
import serial.threaded
from plugwise.constants import (
    BAUD_RATE,
    BYTE_SIZE,
    PARITY,
    SLEEP_TIME,
    STOPBITS,
)
from plugwise.connections.connection import StickConnection
from plugwise.exceptions import PortError
from plugwise.message import PlugwiseMessage
from plugwise.util import PlugwiseException


class Protocol(serial.threaded.Protocol):
    """Serial protocol."""

    def data_received(self, data):
        # pylint: disable-msg=E1101
        self.parser(data)


class PlugwiseUSBConnection(StickConnection):
    """simple wrapper around serial module"""

    def __init__(self, port, stick=None):
        self.port = port
        self.baud = BAUD_RATE
        self.bits = BYTE_SIZE
        self.stop = STOPBITS
        self.parity = serial.PARITY_NONE
        self.stick = stick
        self.run_writer_thread = True
        self.run_reader_thread = True
        self._is_connected = False

    def open_port(self):
        """Open serial port"""
        self.stick.logger.debug("Open serial port")
        try:
            self.serial = serial.Serial(
                port = self.port,
                baudrate = self.baud,
                bytesize = self.bits,
                parity = self.parity,
                stopbits = self.stop,
            )
            self._reader_thread = serial.threaded.ReaderThread(self.serial, Protocol)
            self._reader_thread.start()
            self._reader_thread.protocol.parser = self.feed_parser
            self._reader_thread.connect()
        except serial.serialutil.SerialException as err:
            self.stick.logger.debug(
                "Failed to connect to port %s, %s",
                self.port,
                err,
            )
            raise PortError(err)
        else:
            self.stick.logger.debug("Successfully connected to serial port %s", self.port)
            self._write_queue = Queue()
            self._writer_thread = threading.Thread(None, self.writer_loop,
                                                "write_packets_process", (), {})
            self._writer_thread.daemon = True
            self._writer_thread.start()
            self.stick.logger.debug("Successfully connected to port %s", self.port)
            self._is_connected = True

    def close_port(self):
        """Close serial port."""
        self._is_connected = False
        self.run_writer_thread = False
        try:
            self._reader_thread.close()
        except serial.serialutil.SerialException:
            self.stick.logger.error("Error while closing device")
            raise PlugwiseException("Error while closing device")

    def read_thread_alive(self):
        """Return state of write thread"""
        return self._reader_thread.isAlive()

    def write_thread_alive(self):
        """Return state of write thread"""
        return self._writer_thread.isAlive()

    def is_connected(self):
        """Return connection state"""
        return self._is_connected

    def feed_parser(self, data):
        """Parse received message."""
        assert isinstance(data, bytes)
        self.stick.feed_parser(data)

    def send(self, message, callback=None):
        """Add message to write queue."""
        assert isinstance(message, PlugwiseMessage)
        self._write_queue.put_nowait((message, callback))

    def writer_loop(self):
        """Write thread."""
        while self.run_writer_thread:
            (message, callback) = self._write_queue.get(block=True)
            self.stick.logger.debug("Sending %s to plugwise stick (%s)", message.__class__.__name__, message.serialize())
            self._reader_thread.write(message.serialize())
            time.sleep(SLEEP_TIME)
            if callback:
                callback()


"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise Circle node object
"""
import logging
from datetime import date, datetime, timedelta
from plugwise.constants import (
    SENSOR_AVAILABLE,
    SENSOR_PING,
    SENSOR_POWER_USE,
    SENSOR_POWER_USE_LAST_8_SEC,
    SENSOR_POWER_CONSUMPTION_CURRENT_HOUR,
    SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR,
    SENSOR_POWER_CONSUMPTION_TODAY,
    SENSOR_POWER_CONSUMPTION_YESTERDAY,
    SENSOR_POWER_PRODUCTION_CURRENT_HOUR,
    SENSOR_POWER_PRODUCTION_PREVIOUS_HOUR,
    SENSOR_RSSI_IN,
    SENSOR_RSSI_OUT,
    SWITCH_RELAY,
    HA_SWITCH,
    HA_SENSOR,
    PULSES_PER_KW_SECOND,
)
from plugwise.node import PlugwiseNode

from plugwise.message import PlugwiseMessage
from plugwise.messages.requests import (
    CircleCalibrationRequest,
    CirclePowerBufferRequest,
    CirclePowerUsageRequest,
    CircleSwitchRequest,
)
from plugwise.messages.responses import (
    CircleCalibrationResponse,
    CirclePowerBufferResponse,
    CirclePowerUsageResponse,
    CircleScanResponse,
    CircleSwitchResponse,
)
from plugwise.util import Int


class PlugwiseCircle(PlugwiseNode):
    """provides interface to the Plugwise Circle nodes
    """

    def __init__(self, mac, address, stick):
        super().__init__(mac, address, stick)
        self.categories = (HA_SWITCH, HA_SENSOR)
        self.sensors = (
            SENSOR_AVAILABLE["id"],
            SENSOR_PING["id"],
            SENSOR_POWER_USE["id"],
            SENSOR_POWER_USE_LAST_8_SEC["id"],
            SENSOR_POWER_CONSUMPTION_CURRENT_HOUR["id"],
            SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR["id"],
            SENSOR_POWER_CONSUMPTION_TODAY["id"],
            SENSOR_POWER_CONSUMPTION_YESTERDAY["id"],
            SENSOR_POWER_PRODUCTION_CURRENT_HOUR["id"],
            # SENSOR_POWER_PRODUCTION_PREVIOUS_HOUR["id"],
            SENSOR_RSSI_IN["id"],
            SENSOR_RSSI_OUT["id"],
        )
        self.switches = (SWITCH_RELAY["id"],)
        self.pulses_1s = None
        self.pulses_8s = None
        self.pulses_consumed_1h = None
        self.pulses_produced_1h = None
        self.calibration = False
        self._gain_a = None
        self._gain_b = None
        self._off_noise = None
        self._off_tot = None
        self.power_history = {}
        self.power_consumption_prev_hour = None
        self.power_consumption_today = None
        self.power_consumption_yesterday = None
        self._request_calibration()

    def _request_calibration(self, callback=None):
        """Request calibration info
        """
        self.stick.send(
            CircleCalibrationRequest(self.mac), callback,
        )

    def _request_switch(self, state, callback=None):
        """Request to switch relay state and request state info
        """
        self.stick.send(
            CircleSwitchRequest(self.mac, state), callback,
        )

    def update_power_usage(self, callback=None):
        """Request power usage
        """
        self.stick.send(
            CirclePowerUsageRequest(self.mac), callback,
        )

    def _on_message(self, message):
        """
        Process received message
        """
        if isinstance(message, CirclePowerUsageResponse):
            if self.calibration:
                self._response_power_usage(message)
                self.stick.logger.debug(
                    "Power update for %s, last update %s",
                    self.get_mac(),
                    str(self.last_update),
                )
                self.stick.message_processed(message.seq_id)
            else:
                self.stick.logger.debug(
                    "Received power update for %s before calibration information is known",
                    self.get_mac(),
                )
                self._request_calibration()
        elif isinstance(message, CircleSwitchResponse):
            self._response_switch(message)
            self.stick.logger.debug(
                "Switch update for %s, last update %s",
                self.get_mac(),
                str(self.last_update),
            )
            self.stick.message_processed(message.seq_id)
        elif isinstance(message, CircleCalibrationResponse):
            self._response_calibration(message)
            self.stick.message_processed(message.seq_id)
        elif isinstance(message, CirclePowerBufferResponse):
            if self.calibration:
                self._response_power_buffer(message)
                self.stick.message_processed(message.seq_id)
            else:
                self.stick.logger.debug(
                    "Received power buffer log for %s before calibration information is known",
                    self.get_mac(),
                )
                self._request_calibration()
        else:
            self._circle_plus_message(message)

    def _circle_plus_message(self, message):
        pass

    def _process_scan_response(self, message):
        pass

    def get_relay_state(self) -> bool:
        """ Return last known relay state """
        return self._relay_state

    def set_relay_state(self, state, callback=None):
        """ Switch relay """
        if isinstance(state, bool):
            self._request_switch(state, callback)

    def get_power_usage(self):
        """
        Returns power usage during the last second in Watts
        Based on last received power usage information 
        """
        if self.pulses_1s is None:
            return None
        return self.pulses_to_kWs(self.pulses_1s) * 1000

    def get_power_usage_8_sec(self):
        """
        Returns power usage during the last 8 second in Watts
        Based on last received power usage information 
        """
        if self.pulses_8s is None:
            return None
        return self.pulses_to_kWs(self.pulses_8s, 8) * 1000

    def get_power_consumption_current_hour(self):
        """
        Returns the power usage during this running hour in kWh
        Based on last received power usage information 
        """
        if self.pulses_consumed_1h is None:
            return None
        elif self.pulses_consumed_1h == 0:
            return 0.0
        return self.pulses_to_kWs(self.pulses_consumed_1h, 3600)

    def get_power_production_current_hour(self):
        """
        Returns the power production during this running hour in kWh
        Based on last received power usage information 
        """
        if self.pulses_produced_1h is None:
            return None
        elif self.pulses_produced_1h == 0:
            return 0.0
        return self.pulses_to_kWs(self.pulses_produced_1h, 3600)

    def get_power_consumption_prev_hour(self):
        """ Returns power consumption during the previous hour in kWh
        """
        return self.power_consumption_prev_hour

    def get_power_consumption_today(self):
        """ Total power consumption during today in kWh
        """
        return self.power_consumption_today

    def get_power_consumption_yesterday(self):
        """ Total power consumption of yesterday in kWh
        """
        return self.power_consumption_yesterday

    def _response_switch(self, message):
        """ Process switch response message
        """
        if message.relay_state == b"D8":
            if not self._relay_state:
                self._relay_state = True
                self.do_callback(SWITCH_RELAY["id"])
        else:
            if self._relay_state:
                self._relay_state = False
                self.do_callback(SWITCH_RELAY["id"])

    def _response_power_usage(self, message):
        # sometimes the circle returns max values for some of the pulse counters
        # I have no idea what it means but it certainly isn't a reasonable value
        # so I just assume that it's meant to signal some kind of a temporary error condition

        # Power consumption last second
        if message.pulse_1s.value == 65535:
            self.stick.logger.debug(
                "1 sec power pulse counter for node %s has unreasonable value of 65535",
                self.get_mac(),
            )
        else:
            self.pulses_1s = message.pulse_1s.value
            if message.pulse_1s.value != 0:
                if message.nanosecond_offset.value != 0:
                    pulses_1s = (
                        message.pulse_1s.value
                        * (1000000000 + message.nanosecond_offset.value)
                    ) / 1000000000
                else:
                    pulses_1s = message.pulse_1s.value
                self.pulses_1s = pulses_1s
            else:
                self.pulses_1s = 0
            self.do_callback(SENSOR_POWER_USE["id"])
        # Power consumption last 8 seconds
        if message.pulse_8s.value == 65535:
            self.stick.logger.debug(
                "8 sec power pulse counter for node %s has unreasonable value of 65535",
                self.get_mac(),
            )
        else:
            if message.pulse_8s.value != 0:
                if message.nanosecond_offset.value != 0:
                    pulses_8s = (
                        message.pulse_8s.value
                        * (1000000000 + message.nanosecond_offset.value)
                    ) / 1000000000
                else:
                    pulses_8s = message.pulse_8s.value
                self.pulses_8s = pulses_8s
            else:
                self.pulses_8s = 0
            self.do_callback(SENSOR_POWER_USE_LAST_8_SEC["id"])
        # Power consumption current hour
        if message.pulse_hour_consumed.value == 4294967295:
            self.stick.logger.debug(
                "1 hour consumption power pulse counter for node %s has unreasonable value of 4294967295",
                self.get_mac(),
            )
        else:
            self.pulses_consumed_1h = message.pulse_hour_consumed.value
            self.do_callback(SENSOR_POWER_CONSUMPTION_CURRENT_HOUR["id"])
        # Power produced current hour
        if message.pulse_hour_produced.value == 4294967295:
            self.stick.logger.debug(
                "1 hour power production pulse counter for node %s has unreasonable value of 4294967295",
                self.get_mac(),
            )
        else:
            self.pulses_produced_1h = message.pulse_hour_produced.value
            self.do_callback(SENSOR_POWER_PRODUCTION_CURRENT_HOUR["id"])

    def _response_calibration(self, message):
        """ Store calibration properties
        """
        for x in ("gain_a", "gain_b", "off_noise", "off_tot"):
            val = getattr(message, x).value
            setattr(self, "_" + x, val)
        self.calibration = True

    def pulses_to_kWs(self, pulses, seconds=1):
        """
        converts the amount of pulses to kWs using the calaboration offsets
        """
        if pulses == 0 or not self.calibration:
            return 0.0
        pulses_per_s = pulses / float(seconds)
        corrected_pulses = seconds * (
            (
                (((pulses_per_s + self._off_noise) ** 2) * self._gain_b)
                + ((pulses_per_s + self._off_noise) * self._gain_a)
            )
            + self._off_tot
        )
        calc_value = corrected_pulses / PULSES_PER_KW_SECOND / seconds
        # Fix minor miscalculations
        if calc_value < 0.001 and calc_value > -0.001:
            calc_value = 0.0
        return calc_value

    def _request_power_buffer(self, log_address=None, callback=None):
        """Request power log of specified address
        """
        if log_address == None:
            log_address = self._last_log_address
        if log_address != None:
            if bool(self.power_history):
                # Only request last 2 power buffer logs
                self.stick.send(CirclePowerBufferRequest(self.mac, log_address - 1),)
                self.stick.send(
                    CirclePowerBufferRequest(self.mac, log_address), callback,
                )
            else:
                # Collect power history info of today and yesterday
                # Each request contains 4 hours except last request
                for req_log_address in range(log_address - 13, log_address):
                    self.stick.send(
                        CirclePowerBufferRequest(self.mac, req_log_address),
                    )
                self.stick.send(
                    CirclePowerBufferRequest(self.mac, log_address), callback,
                )

    def _response_power_buffer(self, message):
        """returns information about historical power usage
        each response contains 4 log buffers and each log buffer contains data for 1 hour
        """
        if message.logaddr.value == self._last_log_address:
            self._last_log_collected = True
        # Collect logged power usage
        for i in range(1, 5):
            if getattr(message, "logdate%d" % (i,)).value != None:
                dt = getattr(message, "logdate%d" % (i,)).value
                if getattr(message, "pulses%d" % (i,)).value == 0:
                    self.power_history[dt] = 0.0
                else:
                    self.power_history[dt] = self.pulses_to_kWs(
                        getattr(message, "pulses%d" % (i,)).value, 3600
                    )
        # Cleanup history for more than 2 day's ago
        if len(self.power_history.keys()) > 48:
            for dt in list(self.power_history.keys()):
                if (dt + self.stick.timezone_delta - timedelta(hours=1)).date() < (
                    datetime.now().today().date() - timedelta(days=1)
                ):
                    del self.power_history[dt]
        # Recalculate power use counters
        last_hour_usage = 0
        today_power = 0
        yesterday_power = 0
        for dt in self.power_history:
            if (dt + self.stick.timezone_delta) == datetime.now().today().replace(
                minute=0, second=0, microsecond=0
            ):
                last_hour_usage = self.power_history[dt]
            if (
                dt + self.stick.timezone_delta - timedelta(hours=1)
            ).date() == datetime.now().today().date():
                today_power += self.power_history[dt]
            if (dt + self.stick.timezone_delta - timedelta(hours=1)).date() == (
                datetime.now().today().date() - timedelta(days=1)
            ):
                yesterday_power += self.power_history[dt]
        if self.power_consumption_prev_hour != last_hour_usage:
            self.power_consumption_prev_hour = last_hour_usage
            self.do_callback(SENSOR_POWER_CONSUMPTION_PREVIOUS_HOUR["id"])
        if self.power_consumption_today != today_power:
            self.power_consumption_today = today_power
            self.do_callback(SENSOR_POWER_CONSUMPTION_TODAY["id"])
        if self.power_consumption_yesterday != yesterday_power:
            self.power_consumption_yesterday = yesterday_power
            self.do_callback(SENSOR_POWER_CONSUMPTION_YESTERDAY["id"])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
  name='python-plugwise',
  version="1.0.1",
  author='Frank van Breugel',
  author_email='f.v.breugel@gmail.com',
  description='Async library for Plugwise stick',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/brefra/python-plugwise",
  packages=setuptools.find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Topic :: Home Automation",
  ],
  python_requires='>=3.6',
  install_requires=[
        'crcmod',
        'pyserial',
  ],
)




# Copyright (C) 2011 Sven Petai <hadara@bsd.ee>
# Use of this source code is governed by the MIT license found in the LICENSE file.


class PlugwiseException(Exception):
    """Base error class for this Plugwise library"""

    pass


class PortError(PlugwiseException):
    """Connection to USBstick failed"""

    pass


class StickInitError(PlugwiseException):
    """Initialization of USBstick failed"""

    pass


class NetworkDown(PlugwiseException):
    """Zigbee network not online"""

    pass


class CirclePlusError(PlugwiseException):
    """Connection to Circle+ node failed"""

    pass


class ProtocolError(PlugwiseException):
    """Error while decode received data"""

    pass


class TimeoutException(PlugwiseException):
    """Timeout expired while waiting for response from node"""

    pass



"""
Use of this source code is governed by the MIT license found in the LICENSE file.

Plugwise protocol helpers
"""
import binascii
import crcmod
import datetime
import logging
import re
import struct
import sys
from .exceptions import *
from .constants import *


crc_fun = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0x0000, xorOut=0x0000)


def validate_mac(mac):
    if not re.match("^[A-F0-9]+$", mac):
        return False
    try:
        _ = int(mac, 16)
    except ValueError:
        return False
    return True


def inc_seq_id(seq_id, value=1):
    """
    Increment sequence id by value

    :return: 4 bytes
    """
    temp_int = int(seq_id, 16) + value
    # Max seq_id = b'FFFC'
    if temp_int >= 65532:
        temp_int = 0
    temp_str = str(hex(temp_int)).lstrip("0x").upper()
    while len(temp_str) < 4:
        temp_str = "0" + temp_str
    return temp_str.encode()


class PlugwiseException(Exception):
    """Plugwise Exception."""

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return repr(self.value)


class BaseType(object):
    def __init__(self, value, length):
        self.value = value
        self.length = length

    def serialize(self):
        return bytes(self.value, "utf-8")

    def unserialize(self, val):
        self.value = val

    def __len__(self):
        return self.length


class CompositeType(BaseType):
    def __init__(self):
        self.contents = []

    def serialize(self):
        return b"".join(a.serialize() for a in self.contents)

    def unserialize(self, val):
        for p in self.contents:
            myval = val[: len(p)]
            p.unserialize(myval)
            val = val[len(myval) :]
        return val

    def __len__(self):
        return sum(len(x) for x in self.contents)


class String(BaseType):
    pass


class Int(BaseType):
    def __init__(self, value, length=2):
        self.value = value
        self.length = length

    def serialize(self):
        fmt = "%%0%dX" % self.length
        return bytes(fmt % self.value, "utf-8")

    def unserialize(self, val):
        self.value = int(val, 16)


class UnixTimestamp(Int):
    def __init__(self, value, length=8):
        Int.__init__(self, value, length=length)

    def unserialize(self, val):
        Int.unserialize(self, val)
        self.value = datetime.datetime.fromtimestamp(self.value)


class Year2k(Int):
    """year value that is offset from the year 2000"""

    def unserialize(self, val):
        Int.unserialize(self, val)
        self.value += PLUGWISE_EPOCH


class DateTime(CompositeType):
    """datetime value as used in the general info response
    format is: YYMMmmmm
    where year is offset value from the epoch which is Y2K
    and last four bytes are offset from the beginning of the month in minutes
    """

    def __init__(self, year=0, month=0, minutes=0):
        CompositeType.__init__(self)
        self.year = Year2k(year - PLUGWISE_EPOCH, 2)
        self.month = Int(month, 2)
        self.minutes = Int(minutes, 4)
        self.contents += [self.year, self.month, self.minutes]

    def unserialize(self, val):
        CompositeType.unserialize(self, val)
        minutes = self.minutes.value
        hours = minutes // 60
        days = hours // 24
        hours -= days * 24
        minutes -= (days * 24 * 60) + (hours * 60)
        try:
            self.value = datetime.datetime(
                self.year.value, self.month.value, days + 1, hours, minutes
            )
        except ValueError:
            #debug(
            #    "encountered value error while attempting to construct datetime object"
            #)
            self.value = None


class Time(CompositeType):
    """time value as used in the clock info response"""

    def __init__(self, hour=0, minute=0, second=0):
        CompositeType.__init__(self)
        self.hour = Int(hour, 2)
        self.minute = Int(minute, 2)
        self.second = Int(second, 2)
        self.contents += [self.hour, self.minute, self.second]

    def unserialize(self, val):
        CompositeType.unserialize(self, val)
        self.value = datetime.time(
            self.hour.value, self.minute.value, self.second.value
        )


class IntDec(BaseType):
    def __init__(self, value, length=2):
        self.value = value
        self.length = length

    def serialize(self):
        fmt = "%%0%dd" % self.length
        return bytes(fmt % self.value, "utf-8")

    def unserialize(self, val):
        self.value = val.decode("utf-8")


class RealClockTime(CompositeType):
    """time value as used in the realtime clock info response"""

    def __init__(self, hour=0, minute=0, second=0 ):
        CompositeType.__init__(self)
        self.hour = IntDec(hour, 2)
        self.minute = IntDec(minute, 2)
        self.second = IntDec(second, 2)
        self.contents += [self.second, self.minute, self.hour]

    def unserialize(self, val):
        CompositeType.unserialize(self, val)
        self.value = datetime.time(
            int(self.hour.value),
            int(self.minute.value),
            int(self.second.value),
        )


class RealClockDate(CompositeType):
    """date value as used in the realtime clock info response"""

    def __init__(self, day=0, month=0, year=0):
        CompositeType.__init__(self)
        self.day = IntDec(day, 2)
        self.month = IntDec(month, 2)
        self.year = IntDec(year - PLUGWISE_EPOCH, 2)
        self.contents += [self.day, self.month, self.year]

    def unserialize(self, val):
        CompositeType.unserialize(self, val)
        self.value = datetime.date(
            int(self.year.value) + PLUGWISE_EPOCH,
            int(self.month.value),
            int(self.day.value),
        )


class Float(BaseType):
    def __init__(self, value, length=4):
        self.value = value
        self.length = length

    def unserialize(self, val):
        hexval = binascii.unhexlify(val)
        self.value = struct.unpack("!f", hexval)[0]


class LogAddr(Int):

    def serialize(self):
        return bytes("%08X" % ((self.value * 32) + LOGADDR_OFFSET), "utf-8")

    def unserialize(self, val):
        Int.unserialize(self, val)
        self.value = (self.value - LOGADDR_OFFSET) // 32
